# Original code from http://www.parallelr.com/r-with-parallel-computing/
# Modified by Le Yan

# Not vectorized function
# Quadratic Equation: a*x^2 + b*x + c = 0
solve.quad.eq <- function(a, b, c) 
{
 # Not validate eqution: a and b are almost ZERO
 if(abs(a) < 1e-8 && abs(b) < 1e-8) return(c(NA, NA) )
  
 # Not quad equation
 if(abs(a) < 1e-8 && abs(b) > 1e-8) return(c(-c/b, NA))
  
 # No Solution
 if(b*b - 4*a*c < 0) return(c(NA,NA))
  
 # Return solutions
 x.delta <- sqrt(b*b - 4*a*c)
 x1 <- (-b + x.delta)/(2*a)
 x2 <- (-b - x.delta)/(2*a)
  
 return(c(x1, x2))
}

# Generate data 
len <- 1e7
a <- runif(len, -10, 10)
a[sample(len, 100,replace=TRUE)] <- 0
b <- runif(len, -10, 10)
c <- runif(len, -10, 10)

# serial code
t <- system.time(
res1.s <- lapply(1:len, FUN = function(x) { solve.quad.eq(a[x], b[x], c[x])})
)
print("Serial lapply:")
print(t)

# parallel runs with different number of cores
ncores <- c(2,4,8,16,20)

# parallel, Linux and MAC platform
library(parallel)
# multicores on Linux
for (core in ncores) {
  t <- system.time(
    res1.p <- mclapply(1:len, 
                     FUN = function(x) { solve.quad.eq(a[x], b[x], c[x]) }, 
                     mc.cores = core)
  )
  print(paste("Parallel mclapply with",core,"cores:"))
  print(t)
}

# for style: serial code
res2.s <- matrix(0, nrow=len, ncol = 2)
seqTime <- system.time(
    for(i in 1:len) {
        res2.s[i,] <- solve.quad.eq(a[i], b[i], c[i])
    }
)
print("Serial for:")
print(seqTime)

times <- vector("numeric")
times <- c(times,seqTime["elapsed"])

# foreach, work on Linux/Windows/Mac
library(foreach)
library(doParallel)

for (core in ncores) { 

  cl <- makeCluster(core)
  registerDoParallel(cl, cores=core)
 
  # clusterSplit are very convience to split data but it takes lots of extra memory
  # chunks <- clusterSplit(cl, 1:len)
 
  # split data by ourselves
  chunk.size <- len/core

  t <- system.time(
   res2.p <- foreach(i=1:core, .combine='rbind') %dopar%
   { 
     # local data for results
     res <- matrix(0, nrow=chunk.size, ncol=2)
     for(x in ((i-1)*chunk.size+1):(i*chunk.size)) {
        res[x - (i-1)*chunk.size,] <- solve.quad.eq(a[x], b[x], c[x])
     }
     # return local results
     res
  }
  )
  print(paste("Parallel foreach with",core,"cores:"))
  print(t) 
  times <- c(times,t["elapsed"])

  stopImplicitCluster()
  stopCluster(cl)

}

bm <- data.frame(Worker=c(1,ncores), oneNodeTime=times)

# foreach, distribute workers on different nodes
library(foreach)
library(doParallel)

hosts <- as.vector(unique(read.table(Sys.getenv("PBS_NODEFILE"),stringsAsFactors=F))[,1])
nh <- length(hosts)

times <- seqTime["elapsed"]

for (core in ncores) {

  if (core %% nh != 0) {
    warning("The number of cores must be divisible by the number of nodes!")
    next
  }

  # Create workers on nodes in block manner.
  cl <- makePSOCKcluster(rep(hosts , each = core/nh))
  registerDoParallel(cl, cores=core)

  # clusterSplit are very convience to split data but it takes lots of extra memory
  # chunks <- clusterSplit(cl, 1:len)

  # split data by ourselves
  chunk.size <- len/core

  t <- system.time(
   res2.p <- foreach(i=1:core, .combine='rbind') %dopar%
   {
     # local data for results
     res <- matrix(0, nrow=chunk.size, ncol=2)
     for(x in ((i-1)*chunk.size+1):(i*chunk.size)) {
        res[x - (i-1)*chunk.size,] <- solve.quad.eq(a[x], b[x], c[x])
     }
     # return local results
     res
  }
  )
  print(paste("Parallel foreach with",core,"cores on",nh,"nodes:"))
  print(t)
  times <- c(times,t["elapsed"])

  stopImplicitCluster()
  stopCluster(cl)

}

library(reshape)

bm[,"multiNodeTime"] <- times
colnames(bm) <- c("Worker","On one node","On multiple nodes")
bm.long <- melt(bm,id="Worker",measure=c(2,3))

# Add the ideal values.
bm.long <- rbind.data.frame(bm.long,data.frame(Worker=bm.long$Worker[1:7],variable="Ideal",value=bm.long$value[1]/bm.long$Worker[1:7]))

# Plot the times.

library(ggplot2)

g <- ggplot(bm.long,aes(Worker, value, color=variable))+
  theme_bw()+
  geom_point(data=subset(bm.long,variable!="Ideal"),size=3)+
  geom_line(data=subset(bm.long,variable!="Ideal"),size=2)+
  geom_line(data=subset(bm.long,variable=="Ideal"),size=1,linetype = 2)+
  geom_text(data=subset(bm.long,variable=="On one node"),aes(label=Worker),vjust=-1)+
  ggtitle('Wall clock time vs. number of workers for quadratic equation solver')+
  xlab("Number of workers")+ylab("Wall clock time")+
  xlim(1,50)+ylim(1,200)+
  scale_color_brewer(palette = "Set2")+
  scale_y_log10(breaks=c(100,50,20,10,5))+
  scale_x_log10()+
  theme(plot.title = element_text(size=20),legend.title=element_blank(),text=element_text(size=16))

ggsave(g,filename="R_parallel_benchmark_quadratic.png",width=10,height=8,unit="in")

